#include "Logger.h"
#include <sstream>
#include <algorithm> // std::remove

#include "curl.h" // libcurl, for sending http requests
#include "MD5.h"  // MD5 hashing algorithm
#include "sha1.h" // SHA-1 hashing algorithm
#include "rapidjson\document.h" // json parser

// Each Set function also specifies that the variable is in use.
// That way it won't be added to the event string.
// There might be a more elegant way to do this that I haven't thought of.

void LogEvent::SetArea(std::string area)
{
    area_ = area;
    useArea_ = true;
}

void LogEvent::SetEventID(std::string eventID)
{
    eventID_ = eventID;
    useEventID_ = true;
}

void LogEvent::SetValue(float value)
{
    value_ = value;
    useValue_ = true;
}

void LogEvent::SetLocation(Point location)
{
    location_ = location;
    useLocation_ = true;
}

// Gets the string version of the event, ready to send.
// It might be nicer to use a json generator, but the format is simple enough.
// For reference on what this should like, go here: http://support.gameanalytics.com/entries/22613463-Design-event-structure
std::string LogEvent::GetString(std::string userID, std::string sessionID, std::string build)
{
    std::stringstream ss;
    ss << "{";
    ss << "\"user_id\": ";
    ss << "\"" << userID << "\"";
    ss << ", ";
    ss << "\"session_id\": ";
    ss << "\"" << sessionID << "\"";
    ss << ", ";
    ss << "\"build\": ";
    ss << "\"" << build << "\"";
    if (useArea_)
    {
        ss << ", ";
        ss << "\"area\": ";
        ss << "\"" << area_ << "\"";
    }
    if (useEventID_)
    {
        ss << ", ";
        ss << "\"event_id\": ";
        ss << "\"" << eventID_ << "\"";
    }
    if (useValue_)
    {
        ss << ", ";
        ss << "\"value\": ";
        ss << value_;
    }
    if (useLocation_)
    {
        ss << ", ";
        ss << "\"x\": ";
        ss << location_.x;
        ss << ", ";
        ss << "\"y\": ";
        ss << location_.y;
        ss << ", ";
        ss << "\"z\": ";
        ss << location_.z;
    }
    ss << "}";
    return ss.str();
}

Logger::Logger(void)
{
    // Initialize libcurl.
    curl_global_init(CURL_GLOBAL_ALL);

    // Currently the API's version is 1. Leave it like this for now.
    apiVersion_ = "1";

    // You'll get your game's unique public key from GameAnalytics.
    gameKey_ = "[yourGameKey]";
    // You'll also get an API key from GameAnalytics (for getting heatmaps).
    apiKey_ = "[yourApiKey]";
    // The build version you're on. Set this however works best for you.
    build_ = "[buildVersion]";

    // This will be a concatenation of all events into one string,
    //   enclosed by brackets.
    gaEvents_ = "[";
}

// Although it's necessary to include a user ID and session ID in an event,
//   there's no regulation on what that entails. I use the following two
//   functions to ensure that the user ID is unique to each machine the
//   game is played on, and the session ID is unique every time.
// More info here: http://support.gameanalytics.com/entries/23054568-Generating-unique-user-identifiers

// This hashes the user's MAC address to create an ID unique to that machine.
void Logger::SetUserID(void)
{
    // Prepare an array to get info for up to 16 network adapters.
    // It's a little hacky, but it's way complicated to do otherwise,
    //   and the chances of someone having more than 16 are pretty crazy.
    IP_ADAPTER_INFO info[16];

    // Get info for all the network adapters.
    DWORD dwBufLen = sizeof(info);
    DWORD dwStatus = GetAdaptersInfo(info, &dwBufLen);
    if (dwStatus != ERROR_SUCCESS)
    {
        /* deal with error */
    }
    else
    {
        PIP_ADAPTER_INFO pAdapterInfo = info;

        // Iterate through the adapter array until we find one.
        // (Will most likely be the first.)
        while (pAdapterInfo && pAdapterInfo == 0)
        {
            pAdapterInfo = pAdapterInfo->Next;
        }
        
        if (!pAdapterInfo)
        {
            // No network adapter found. In my code I just use a GUID at this point,
            //   (same code to get a session ID), but I'll leave that up to you.
        }
        else
        {
            // Get a hash of the MAC address using SHA1.
            unsigned char hash[10];
            char hexstring[41];
            sha1::calc(pAdapterInfo->Address, pAdapterInfo->AddressLength, hash);
            sha1::toHexString(hash, hexstring);
            userID_ = hexstring;

            // Remove any hyphens.
            userID_.erase(std::remove(userID_.begin(), userID_.end(), '-'), userID_.end());
        }
    }
}

// Sets the session ID with a GUID (Globally Unique Identifier).
void Logger::SetSessionID(void)
{
    // Get the GUID.
    GUID guid;
    CoCreateGuid(&guid);

    // Turn the GUID into an ASCII string.
    RPC_CSTR str;
    UuidToStringA((UUID*)&guid, &str);

    // Get a standard string out of that.
    session_id_ = (LPSTR)str;

    // Remove any hyphens.
    session_id_.erase(std::remove(session_id_.begin(), session_id_.end(), '-'), session_id_.end());
}

// Adds an event to the string of all events.
// This allows you to decide when to send events to the server,
//   as opposed to sending one every time an event is logged.
void Logger::AddLogEvent(LogEvent ev)
{
    // If this isn't the first event...
    if (gaEvents_.back() != '[')
    {
        // ...add a comma to separate this one from the last.
        gaEvents_ += ", ";
    }
    // Add the event to the string.
    gaEvents_ += ev.GetString(userID_, sessionID_, build_);
}

// Submits the events string to the server.
// This is the meat of the whole operation.
// (I think I mixed up two phrases there...)
void Logger::SubmitLogEvents(void)
{
    // New CURL object.
    CURL * curl = curl_easy_init();
    if (curl)
    {
        // Close the events string with a bracket.
        gaEvents_ += "]";

        // There are multiple categories of events, but I only use "design" so far.
        // To use different categories you would likely have to rearrange this system,
        //   perhaps by having different events strings for different categories.
        std::string category = "design";

        // The URL to send events to
        std::string url = "http://api.gameanalytics.com/" + apiVersion_ + "/" + gameKey_ + "/" + category;

        // You'll get your secret key from GameAnalytics.
        // You could store it as a member variable but I like to keep it
        //   in memory as little as possible. I don't know if that matters,
        //   but whatever.
        std::string secretKey = "[yourSecretKey]";

        // Combine your events string and secret key for the header.
        std::string header = gaEvents_ + secretKey;
        // Then hash that, and give it the label "Authorization:".
        std::string auth = "Authorization:" + md5(header);

        // Create a struct of headers.
        struct curl_slist *chunk = NULL;
        // Add the auth header to the struct.
        chunk = curl_slist_append(chunk, auth.c_str());

        // Set the URL of the request.
        curl_easy_setopt(curl, CURLOPT_URL, url);
        // Set the POST data (the string of events).
        curl_easy_setopt(curl, CURLOPT_POSTFIELDS, gaEvents_.c_str());
        // Set the headers (hashed secret key + data).
        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, chunk);
        // Set request mode to POST.
        curl_easy_setopt(curl, CURLOPT_POST, 1);

        // Send the request!
        CURLcode res = curl_easy_perform(curl);

        // Check the error flag.
        if (res != CURLE_OK)
        {
            // This is how I roll with error messages.
            std::cout << "Curl didn't not even work. " << curl_easy_strerror(res) << std::endl;
        }

        // Cleanup your CURL object.
        curl_easy_cleanup(curl);
        // Free the headers.
        curl_slist_free_all(chunk);
    }

    // Start the events string over again.
    gaEvents_ = "[";
}

// Get heatmap data for an area, and add it to your collection of heatmaps.
void Logger::LoadHeatmap(std::string area, std::string eventID)
{
    // New CURL object.
    CURL * curl = curl_easy_init();
    if (curl)
    {
        // Request URL for getting heatmap data.
        std::string url = "http://data-api.gameanalytics.com/heatmap";
        // Specify what area the heatmap should be from, and what events it
        //   should be getting data for.
        // Note: you can ask for multiple events at the same time with |,
        //   like "Death:Enemy|Death:Fall".
        std::string requestInfo = "game_key=" + gameKey_ + "&event_ids=" + eventID + "&area=" + area;
        // Concatenate the request with the URL.
        url += "/?";
        url += requestInfo;

        // Build the header with the request and API key.
        std::string header = requestInfo + apiKey_;
        // Hash that, and give it the label "Authorization:".
        std::string auth = "Authorization:" + md5(header);

        // Create a struct of headers.
        struct curl_slist *chunk = NULL;
        // Add the auth header to the struct.
        chunk = curl_slist_append(chunk, auth.c_str());

        // Set the URL of the request.
        curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
        // Set the headers (hashed API key + request).
        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, chunk);
        // Set the callback to deal with the returned data.
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, DataToString);

        std::string result;

        // Pass in a string to accept the result.
        // (This will be taken care of in the callback function.)
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, &result);

        // Send the request!
        CURLcode res = curl_easy_perform(curl);

        // Check the error flag.
        if (res != CURLE_OK)
        {
            // Bummer.
            std::cout << "Curl didn't not even work. " << curl_easy_strerror(res) << std::endl;
        }

        // Get ready to receive json data.
        rapidjson::Document doc;

        // Parse json data.
        if(doc.Parse<0>(result.c_str()).HasParseError())
        {
            /* deal with error */
        }

        // Data returned consists of 4 arrays: x, y, z, and value.
        // x, y, and z being coordinates of course, and value being
        //   the number of instances at that point.
        rapidjson::Value & x = doc["x"];
        rapidjson::Value & y = doc["y"];
        rapidjson::Value & z = doc["z"];
        rapidjson::Value & value = doc["value"];

        // Make a deque of pairs: points and their values.
        std::deque<std::pair<Point, int>> points;

        // Iterate through the arrays. (We're just iterating through x technically,
        //   but all arrays are the same length.)
        for (unsigned i = 0; i < x.Size(); ++i)
        {
            // Add x, y, and z as a Point, and value as the second part of the pair.
            points.push_back(std::pair<Point, int>(Point(x[i].GetDouble(), y[i].GetDouble(), z[i].GetDouble()), value[i].GetInt()));
        }

        // I like to keep my heatmaps ordered by area.
        // You can organize them however makes sense to you.
        heatmaps_.insert(std::pair<std::string, std::deque<std::pair<Point, int>>>(area, points));
        
        // Cleanup your CURL object.
        curl_easy_cleanup(curl);
        // Free the headers.
        curl_slist_free_all(chunk);
    }
}

// Gets a heatmap from a particular area. If that heatmap doesn't exist, returns false.
bool Logger::GetHeatmap(std::string area, std::deque<std::pair<Point, int>>& points)
{
    std::map<std::string, std::deque<std::pair<Point, int>>>::iterator it = heatmaps_.find(area);
    if (it == heatmaps_.end())
        return false;

    points = *it;
    return true;
}

// Callback to take returned data and convert it to a string.
// The last parameter is actually user-defined - it can be whatever
//   type you want it to be.
int Logger::DataToString(char *data, size_t size, size_t nmemb, std::string * result)
{
    int count = 0;
    if (result)
    {
        // Data is in a character array, so just append it to the string.
        result->append(data, size * nmemb);
        count = size * nmemb;
    }
    return count;
}
